from collections import defaultdict
from swagger_schema import Swagger, Info, Path
from web_transmute.swagger import (
    generate_swagger,
    get_swagger_static_root
)
import tornado.web

METHODS = ["get", "post", "delete", "put"]
STATIC_ROOT = "/_swagger/static"


def generate_swagger_json(app, title="swagger"):
    """
    from a tornado application, generate a swagger dict
    for all of the transmute routes available.
    """
    paths = defaultdict(Path)
    for domain, specs in app.handlers:
        for s in specs:
            path = s.regex.pattern
            handler = s.handler_class
            for m in METHODS:
                method = getattr(handler, m)
                if hasattr(method, "transmute_func"):
                    setattr(paths[path], m, method.transmute_func.swagger)
    return Swagger(
        info=Info(title=title, version="1.0"),
        paths=paths,
        swagger="2.0"
    ).dump()


def generate_swagger_json_handler(app, title="swagger"):

    swagger_json = generate_swagger_json(app, title=title)

    class SwaggerSpecHandler(tornado.web.RequestHandler):

        def get(self):
            self.write(swagger_json)
            self.finish()

    return SwaggerSpecHandler


def add_swagger_static_routes(app, target_route, swagger_json_route):
    static_root = get_swagger_static_root()
    swagger_body = generate_swagger(STATIC_ROOT, swagger_json_route)

    class SwaggerBodyHandler(tornado.web.RequestHandler):

        def get(self):
            self.write(swagger_body)
            self.finish()

    app.add_handlers(".*", [
        (target_route, SwaggerBodyHandler),
        (STATIC_ROOT + "/(.*)", tornado.web.StaticFileHandler, {"path": static_root})
    ])



import tornado.web
from web_transmute import default_context
from .parameters import get_param_extractor


class HandlerFactory(object):
    """
    HandleFactory handles aggregating transmute functions
    into a set of handlers.
    """

    def __init__(self):
        self._handlers = {}

    def add_handler(self, path, transmute_func):
        if path not in self._handlers:
            self._handlers[path] = _create_handler()

        handler = self._handlers[path]

        _add_transmute_func_to_handler(transmute_func, handler)

    def get_handler_tuples(self):
        handlers = []
        for path, handler in self._handlers.items():
            handlers.append((path, handler))
        return handlers


def _create_handler():

    class Handler(tornado.web.RequestHandler):
        pass

    return Handler


def _add_transmute_func_to_handler(transmute_func, handler,
                                   context=default_context):
    handler_method = generate_handler_method(transmute_func)

    is_get = True

    for method in transmute_func.http_methods:
        is_get = False
        setattr(handler, method.lower(), handler_method)

    if is_get:
        setattr(handler, "get", handler_method)


def generate_handler_method(transmute_func, context=default_context):

    extract_args = get_param_extractor(transmute_func, context)

    def method(self, *args):
        try:
            args, kwargs = extract_args(self, args=args)
            result = transmute_func.raw_func(self, *args, **kwargs)
            self.set_header("Content-Type", "application/json")
            self.write({
                "success": True,
                "result": result
            })
        except Exception as e:
            if (transmute_func.error_exceptions and
               isinstance(e, transmute_func.error_exceptions)):
                self.write({
                    "success": False,
                    "details": str(e)
                })
            else:
                raise
    method.transmute_func = transmute_func

    return method



from web_transmute import *
from web_transmute.function import TransmuteFunction
from .handler import generate_handler_method
from .swagger import (
    generate_swagger_json,
    generate_swagger_json_handler,
    add_swagger_static_routes
)


def to_route(**options):

    def decorator(fn):
        transmute_func = TransmuteFunction(fn, **options)
        return generate_handler_method(transmute_func)

    return decorator



import tornado.testing
import tornado.web
import tornado_transmute
import json


def _create_app():
    app = tornado.web.Application([
        ("/foo/([^\/]+)", ExampleHandler)
    ])
    swagger_handler = tornado_transmute.generate_swagger_json_handler(app)
    tornado_transmute.add_swagger_static_routes(
        app, "/swagger", "/swagger.json"
    )
    app.add_handlers(".*", [("/swagger.json", swagger_handler)])
    return app


class TestApp(tornado.testing.AsyncHTTPTestCase):

    def get_app(self):
        return _create_app()

    def test_foo(self):
        resp = self.fetch("/foo/bar")
        assert resp.code == 200
        resp_json = json.loads(resp.body.decode("UTF-8"))
        assert resp_json.get("success")
        assert resp_json.get("result") == 2

    def test_foo_queryparam(self):
        resp = self.fetch("/foo/bar?multiplier=2")
        assert resp.code == 200
        resp_json = json.loads(resp.body.decode("UTF-8"))
        assert resp_json.get("success")
        assert resp_json.get("result") == 4

    def test_swagger_json(self):
        resp = self.fetch("/swagger.json")
        assert resp.code == 200
        resp_json = json.loads(resp.body.decode("UTF-8"))
        assert "get" in resp_json["paths"]["/foo/([^\/]+)$"]
        assert resp_json["swagger"] == "2.0"
        assert resp_json["info"] == {
            "title": "swagger", "version": "1.0"
        }

    def test_swagger_body(self):
        resp = self.fetch("/swagger")
        assert resp.code == 200
        body = resp.body.decode("UTF-8")
        assert "swagger-section" in body


class ExampleHandler(tornado.web.RequestHandler):

    @tornado_transmute.to_route()
    @tornado_transmute.annotate({
        "resource": str, "multiplier": int,
        "return": int
    })
    def get(self, resource, multiplier=1):
        return 2 * multiplier

if __name__ == "__main__":
    app = _create_app()
    app.listen(8888)
    tornado.ioloop.IOLoop.current().start()



from functools import partial


def get_param_extractor(transmute_func, context):
    if "GET" in transmute_func.http_methods:
        return partial(_extract_from_queryparams, transmute_func, context)
    else:
        return partial(_extract_from_body, transmute_func, context)


def _extract_from_queryparams(transmute_func, context, handler, args=None):
    args, kwargs = args or [], {}
    # TODO: there needs to be a good strategy to parse out
    # positional arguments keyward arguments
    # (due to tornado accepting positional arguments matching regex.)
    index = len(args)
    for arg in transmute_func.signature.args[index:]:
        value = handler.get_query_argument(arg.name)
        args.append(context.serializers.load(arg.type, value))

    for name, info in transmute_func.signature.kwargs.items():
        value = handler.get_query_argument(name, default=None)
        if value is not None:
            kwargs[name] = context.serializers.load(info.type, value)

    return args, kwargs


def _extract_from_body(transmute_func, context, handler, args=None):
    kwargs = {}
    # TODO: there needs to be a good strategy to parse out
    # positional arguments keyword arguments
    # (due to tornado accepting positional arguments matching regex.)
    for name, info in transmute_func.arguments.items():
        serializer = context.serializer[info.type]
        value = handler.get_query_argument(name, default=None)
        if value is not None:
            kwargs[name] = serializer.deserialize(value)
    return kwargs
